/* $Id: blued.c,v 1.2 1999/12/28 11:56:29 kjc Exp kjc $ */
/*
 * Copyright (C) 1997, 1998
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <syslog.h>

#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <net/if.h>

#include <altq/altq_blue.h>

#define BLUE_DEV	"/dev/altq/blue"

int fd = -1;
int debug = 0;
int use_ecn = 1;
int config = 0;
int maxqlen = 20;
int bandwidth = 10000000;	/* 10Mbps */
int packet_size = 1000;		/* average packet size: 1000 bytes */
int blue_max_pmark = 4000;
int blue_hold_time = 1000;

static void usage(void)
{
    fprintf(stderr, "usage: blued [-de] [-l maxqlen] [-h hold_time] [-b bandwidth] [-s packet_size] if_name\n");
    fprintf(stderr, "		-d: debug\n");
    fprintf(stderr, "		-e: use ECN (packet marking)\n");
    fprintf(stderr, "		-l maxqlen: set max queue length\n");
    fprintf(stderr, "		-b bandwidth\n");
    fprintf(stderr, "		-h holdtime\n");
    fprintf(stderr, "		-s packet_size\n");
    exit(1);
}

static void
sigint_handler(int sig)
{
    /* do nothing */
}

int
main(argc, argv)
	int argc;
	char *argv[];
{
    int ch;
    struct blue_interface iface;
    char *if_name, *cp;

    while ((ch = getopt(argc, argv, "de:l:h:b:p:s:")) != EOF) {
	switch (ch) {
	case 'd':
	    debug = 1;
	    break;
	case 'e':
	    use_ecn = 1;
	    config = 1;
	    break;
	case 'l':
	    maxqlen = atoi(optarg);
	    config = 1;
	    break;
	case 'p':
	    blue_max_pmark = atoi(optarg);
	    config = 1;
	    break;
	case 'h':
	    blue_hold_time = atoi(optarg);
	    config = 1;
	    break;
	case 'b':
	    cp = NULL;
	    bandwidth = strtol(optarg, &cp, 0);
	    if (cp != NULL) {
		if (*cp == 'K' || *cp == 'k')
		    bandwidth *= 1000;
		if (*cp == 'M' || *cp == 'm')
		    bandwidth *= 1000000;
	    }
	    config = 1;
	    break;
	case 's':
	    cp = NULL;
	    packet_size = strtol(optarg, &cp, 0);
	    if (cp != NULL && (*cp == 'K' || *cp == 'k'))
		    packet_size *= 1024;
	    config = 1;
	    break;
	}
    }
    argc -= optind;
    argv += optind;
    if (argc > 0)
	if_name = argv[0];
    else {
	if (argc == 0)
	    fprintf(stderr, "missing interface name\n");
	usage();
    }

    signal(SIGINT, sigint_handler);
    signal(SIGTERM, sigint_handler);

    openlog("blued", (debug ? (LOG_PID|LOG_PERROR) : LOG_PID), LOG_DAEMON);

    if (!debug)
	daemon(0, 0);

    if((fd = open(BLUE_DEV, O_RDWR)) == NULL) {
	syslog(LOG_ERR, "can't open %s: %m", BLUE_DEV);
	exit(1);
    }

    if (debug) {
	printf("blue on interface %s\n", if_name);
	printf("  set queue limit to %d\n", maxqlen);
	printf("  set maxpmark to %d\n", blue_max_pmark);
	printf("  set hold_time to %d\n", blue_hold_time);
	printf("  set bandwidth to %d bps\n", bandwidth);
	printf("  set packet_size to %d bytes\n", packet_size);
    }

    iface.blue_ifname[IFNAMSIZ-1] = '\0';
    strncpy(iface.blue_ifname, if_name, IFNAMSIZ-1);
    
    if (ioctl(fd, BLUE_IF_ATTACH, &iface) < 0) {
	syslog(LOG_ERR, "BLUE_IF_ATTACH for %s: %m", if_name);
	exit(1);
    }

    if (config) {
	struct blue_conf conf;

	bzero(&conf, sizeof(struct blue_conf));
	conf.iface = iface;
	conf.blue_limit = maxqlen;
	conf.blue_pkttime = packet_size * 8 * 1000 / (bandwidth / 1000);
	conf.blue_flags = use_ecn ? BLUEF_ECN : 0;
	conf.blue_max_pmark = blue_max_pmark;
	conf.blue_hold_time = blue_hold_time;
	
	if (debug)
	    printf("  packet time set to %d usec\n", conf.blue_pkttime);
	    
	if (ioctl(fd, BLUE_CONFIG, &conf) < 0) {
	    syslog(LOG_ERR, "BLUE_CONFIG: %m");
	    exit(1);
	}
    }

    if (ioctl(fd, BLUE_ENABLE, &iface) < 0) {
	syslog(LOG_ERR, "BLUE_ENABLE: %m");
	exit(1);
    }

    /* wait for signals and then terminate */
    pause();

    if (debug)
	printf("Got signal.  Exiting...\n");

    if (ioctl(fd, BLUE_DISABLE, &iface) < 0)
	syslog(LOG_ERR, "BLUE_DISABLE: %m");

    if (ioctl(fd, BLUE_IF_DETACH, &iface) < 0)
	syslog(LOG_ERR, "BLUE_IF_DETACH: %m");

    close(fd);
    closelog();

    return 0;
}
