/*
 * Copyright (C) 1997
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: cbqprobe.c,v 0.8 1999/12/28 11:56:46 kjc Exp kjc $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <signal.h>
#include <limits.h>
#include <errno.h>
#include <varargs.h>

#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/time.h>

#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <altq/altq.h>
#include <altq/altq_cbq.h>

#include "cbqmon.h"

#define	CBQ_DEVICE	"/dev/altq/cbq"
#define NCLASSES	64

int cbqfd = -1;
int sockfd = -1;
union sockunion view_addr;
int family = AF_UNSPEC;

char *if_name = NULL;
int interval = 250;	/* msec */

char buffer[sizeof(struct cbqm_info)];

char *dumpfile = NULL;
FILE *dumpf = NULL;
int replay_speed = 0;

static void replay_dumpfile(FILE *);
static void endian_swap(char *buf, int size);

static void usage(void)
{
    fprintf(stderr, "usage: cbqprobe interface [options] host\n");
    fprintf(stderr, "       options:\n");
    fprintf(stderr, "           [-i interval_ms]\n");
    fprintf(stderr, "           [-p dst_port]\n");
    fprintf(stderr, "           [-a local_address]\n");
    fprintf(stderr, "           [-t ttl]\n");
    fprintf(stderr, "           [-4|-6]: use ipv4/ipv6\n");
    fprintf(stderr, "           [-r dumpfile [-s replay_speed]]\n");
    exit(1);
}

void fatal_error(va_alist)
    va_dcl
{
    va_list ap;
    char *fmt;

    if (errno != 0)
	perror("fatal_error");
    else
	fprintf(stderr, "fatal_error: ");
    va_start(ap);
    fmt = va_arg(ap, char *);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    fprintf(stderr, "\n");
    exit(1);
}

static void
sigint_handler(int sig)
{
    fprintf(stderr, "Exiting on signal %d\n", sig);

    if (cbqfd != -1)
	close(cbqfd);
    if (sockfd != -1)
	close(sockfd);
    if (dumpf != NULL)
	fclose(dumpf);

    exit(0);
}

void sleep_ms(int ms)
{
    struct timeval wait_time;

    wait_time.tv_sec = ms / 1000;
    wait_time.tv_usec = (ms % 1000) * 1000;
    select(0, NULL, NULL, NULL, &wait_time);
}

/* convert a string name to struct sockaddr_in */
int name2sockaddr(char *name, int port, union sockunion *addrp, int family)
{
#ifdef INET6
    struct addrinfo hints, *res;
    char portstr[64];
    int error;

    sprintf(portstr, "%d", port);
    memset(&hints, 0, sizeof(struct addrinfo));
    hints.ai_family = family;
    hints.ai_socktype = SOCK_DGRAM;
    if (name == NULL)
	hints.ai_flags = AI_PASSIVE;

    if ((error = getaddrinfo(name, portstr, &hints, &res)) != 0) {
	fprintf(stderr, "can't get addrinfo for %s: %s\n",
		name, gai_strerror(error));
	return (-1);
    }

    *addrp = *(union sockunion *)res->ai_addr;

    freeaddrinfo(res);
#else /* INET6 */    
    unsigned long inaddr;
    struct hostent *hep;

    bzero(addrp, sizeof(union sockunion));
    addrp->su_sin.sin_family = AF_INET;
    addrp->su_sin.sin_len = sizeof(struct sockaddr_in);
    if (name != NULL) {
	if ((inaddr = inet_addr(name)) != -1)
	    bcopy(&inaddr, &addrp->su_sin.sin_addr, sizeof(inaddr));
	else if ((hep = gethostbyname(name)) != NULL)
	    bcopy(hep->h_addr, &addrp->su_sin.sin_addr, hep->h_length);
	else
	    return (-1);
    }
    else
	addrp->su_sin.sin_addr.s_addr = htonl(INADDR_ANY);
    addrp->su_sin.sin_port = htons(port);
#endif /* INET6 */    
    return 0;
}

int main(int argc, char **argv)
{
    struct cbq_getstats	   	get_stats;
    class_stats_t		*sp, stats[NCLASSES];
    struct timeval		cur_time;
    struct cbqm_info		*info;
    struct class_info		*cinfo;
    union sockunion my_addr;
    u_char ttl = 1;			/* time-to-live field for mcast */
    int port_no = CBQM_PORT;		/* receiver's port number */
    int multicast = 0;	/* use multicast */
    char *view_name = "linus";
    char *my_name = NULL;
    int				i, len, rval;

    if (argc < 2)
	usage();

    if_name = argv[1];

    argc -= 2;
    argv += 2;
    while(argc-- > 0) {
	if (argv[0][0] == '-') {
	    switch (argv[0][1]) {
	    case 'i':
		--argc;
		interval = atoi(*++argv);
		printf("interval set to %d msec.\n", interval);
		break;
	    case 'a':
		--argc;
		my_name = *++argv;
		printf("local host is %s\n", my_name);
		break;
	    case 'r':
		--argc;
		dumpfile = *++argv;
		printf("reading dump file %s\n", dumpfile);
		break;
	    case 'p':
		--argc;
		port_no = atoi(*++argv);
		printf("dest port is %d\n", port_no);
		break;
	    case 's':
		--argc;
		replay_speed = atoi(*++argv);
		printf("replay speed is %d\n", replay_speed);
		break;
	    case 't':
		--argc;
		ttl = atoi(*++argv);
		printf("ttl is %d\n", ttl);
		break;
	    case '4':
		family = AF_INET;
		break;
	    case '6':
		family = AF_INET6;
		break;
	    default:
		usage();
		break;
	    }
	}
	else {
	    view_name = *argv;
	    printf("viewer is %s\n", view_name);
	}
	++argv;
    }

    if (if_name == NULL || view_name == NULL)
	usage();

    signal(SIGINT, sigint_handler);
    signal(SIGTERM, sigint_handler);

    /* initialize socket */
    if (name2sockaddr(view_name, port_no, &view_addr, family) < 0)
	fatal_error("can't get viewer's address!");

    if (name2sockaddr(my_name, 0, &my_addr, view_addr.su_family) < 0)
	fatal_error("can't get my address!");

#ifdef IN_MULTICAST
    if (view_addr.su_family == AF_INET &&
	IN_MULTICAST(ntohl(view_addr.su_sin.sin_addr.s_addr)))
	multicast = 4;
#endif
#ifdef INET6
    if (view_addr.su_family == AF_INET6 &&
	IN6_IS_ADDR_MULTICAST(&view_addr.su_sin6.sin6_addr))
	multicast = 6;
#endif

    if ((sockfd = socket(my_addr.su_family, SOCK_DGRAM, 0)) < 0)
	fatal_error("sender: can't open socket");

#ifdef IP_MULTICAST_TTL
    if (multicast == 4 && ttl != 1)
	if (setsockopt(sockfd, IPPROTO_IP, IP_MULTICAST_TTL,
		       &ttl, sizeof(ttl)) < 0)
	    fatal_error("can't set ttl");
#endif
#ifdef INET6
    if (multicast == 6 && ttl != 1) {
	int hops = ttl;
	if (setsockopt(sockfd, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
		       &hops, sizeof(hops)) < 0)
	    fatal_error("can't set ttl");
    }
#endif

    if (bind(sockfd, (struct sockaddr *)&my_addr, my_addr.su_len) < 0)
	fatal_error("sender: can't bind");

    if (dumpfile != NULL) {
	/* if dumpfile is specified, open dumpfile */
	if ((dumpf = fopen(dumpfile, "r")) == NULL)
	    fatal_error("can't open dumpfile %s!", dumpfile);
	replay_dumpfile(dumpf);
	/* never returns */
    }

    /* open cbq */
    if ((cbqfd = open(CBQ_DEVICE, O_RDONLY)) < 0) {
	perror("cbq open");
	exit(1);
    }

    if (strlen(if_name) > IFNAMSIZ) {
	fprintf(stderr, "interface name %s too long!\n", if_name);
	exit(1);
    }
	
    strcpy(get_stats.iface.cbq_ifacename, if_name);
    get_stats.stats = stats;

    info = (struct cbqm_info *)buffer;
    info->cbqm_magic = CBQM_MAGIC;
    
    while (1) {

	get_stats.nclasses = NCLASSES;
	
	if (ioctl(cbqfd, CBQ_GETSTATS, &get_stats) < 0) {
	    perror("ioctl CBQ_GETSTATS");
	    exit(1);
	}

	gettimeofday(&cur_time, NULL);

	info->cbqm_sec = cur_time.tv_sec;
	info->cbqm_usec = cur_time.tv_usec;

	for (i=0, cinfo=&info->cbqm_cinfo[0]; i<get_stats.nclasses;
	     i++, cinfo++) {
	    sp = &stats[i];
	    cinfo->handle = sp->handle;
	    cinfo->npackets = sp->xmit_cnt.packets;
	    cinfo->drops = sp->drop_cnt.packets;
	    cinfo->nbytes = (u_int)sp->xmit_cnt.bytes;
	    cinfo->drop_bytes = (u_int)sp->drop_cnt.bytes;
	}

	info->cbqm_len = i;
	len = sizeof(struct cbqm_info) - (CBQM_MAXCLASS - i)*sizeof(struct class_info);

	if ((rval = sendto(sockfd, buffer, len, 0,
		       (struct sockaddr *)&view_addr, view_addr.su_len)) < 0)
	    perror("sendto");

	sleep_ms(interval);
    }
}

static void replay_dumpfile(FILE *df)
{
    struct cbqm_info *info = (struct cbqm_info *)buffer;
    int len, hdr_len, interval, rval;
    struct timeval last_time;
    int swap = 0;
    
    last_time.tv_sec = 0;
    
    hdr_len = sizeof(struct cbqm_info)
	- CBQM_MAXCLASS * sizeof(struct class_info);
    
    while (1) {
	/* read the header */
	if (fread(buffer, hdr_len, 1, df) != 1) {
	    if (feof(df)) {
		printf("end-of-dumpfile\n");
		sleep(1);
		exit(0);
	    }
	    fatal_error("can't read dumpfile header!");
	}

	if (info->cbqm_magic != CBQM_MAGIC) {
	    if (info->cbqm_magic == CBQM_CIGAM) {
		swap = 1;
		endian_swap(buffer, hdr_len);
	    }
	    else {
		fatal_error("bad magic! 0x%x != 0x%x",
			    info->cbqm_magic, CBQM_MAGIC);
	    }
	}

	/* read the rest of the record */
	if (fread(buffer+hdr_len, sizeof(struct class_info),
		  info->cbqm_len, df)
	    != info->cbqm_len)
	    fatal_error("can't read dumpfile header!");
	if (swap)
	    endian_swap(buffer + hdr_len,
			info->cbqm_len*sizeof(struct class_info));
	
	len = hdr_len + info->cbqm_len * sizeof(struct class_info);

	if (last_time.tv_sec == 0)
	    /* first record */
	    interval = 0;
	else {
	    long us;
	    interval = (info->cbqm_sec - last_time.tv_sec) * 1000;
	    us = (info->cbqm_usec - last_time.tv_usec);
	    while (us < 0) {
		interval -= 1;
		us += 1000;
	    }
	    interval += us / 1000;
	}

	last_time.tv_sec = info->cbqm_sec;
	last_time.tv_usec = info->cbqm_usec;

	if (replay_speed > 0)
	    interval = interval /replay_speed;
	
	sleep_ms(interval);

	if ((rval = sendto(sockfd, buffer, len, 0,
		    (struct sockaddr *)&view_addr, view_addr.su_len)) < 0)
	    perror("sendto");

    }
}

/* do endian conversion.  we assume all fields are 32 bits. */
static void endian_swap(char *buf, int size)
{
    long tmp, *p = (long *)buf;
    char *cp, *tp = (char *)&tmp;
    int i;
    
    cp = buf;
    for (i=0; i<size; i+=4) {
	tmp = *p++;
	cp[0] = tp[3]; cp[1] = tp[2]; cp[2] = tp[1]; cp[3] = tp[0];
	cp += 4;
    }
}
