/*
 * Copyright (C) 1999
 *	Sony Computer Science Laboratories, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: cdnrstat.c,v 1.2 1999/12/28 11:57:09 kjc Exp kjc $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <signal.h>
#include <errno.h>
#include <err.h>

#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/time.h>

#include <sys/socket.h>
#include <net/if.h>

#include <netinet/in.h>
#include <altq/altq.h>
#include <altq/altq_cdnr.h>

#define	CDNR_DEVICE	"/dev/altq/cdnr"
#define NELEMENTS	64
#define MAX_PROB	(128*1024)

/*
 * cast u_int64_t to ull for printf, since type of u_int64_t
 * is architecture dependent
 */
typedef	unsigned long long	ull;

char *element_names[] = { "none", "top", "element", "tbmeter", "trtcm" };
char *tbmprof_names[] = { "in:    ", "out:   " };
char *tcmprof_names[] = { "green: ", "yellow:", "red:   " };
int cdnrfd = -1;

char *if_name = NULL;
int interval = 5;
struct tce_stats	elements[NELEMENTS], last[NELEMENTS];

static void usage(void)
{
	fprintf(stderr, "usage: cdnrstat interface [-i interval]\n");
	exit(1);
}

static void
sigint_handler(int sig)
{
	fprintf(stderr, "Exiting on signal %d\n", sig);

	close(cdnrfd);

	exit(0);
}

int main(int argc, char **argv)
{
	struct cdnr_get_stats	get_stats;
	struct tce_stats	*tce, *lp;
	struct timeval		cur_time, last_time;
	double			sec, mbps;
	char			**profile_names;
	int			i, j, nprofile;

	if (argc < 2)
		usage();

	if_name = argv[1];

	if (argc == 4) {
		if (strncmp(argv[2], "-i", 2) == 0)
			interval = atoi(argv[3]);
		else
			usage();
	}

	signal(SIGINT, sigint_handler);
	signal(SIGTERM, sigint_handler);
    
	if ((cdnrfd = open(CDNR_DEVICE, O_RDONLY)) < 0) {
		perror("cdnr open");
		exit(1);
	}

	if (strlen(if_name) > IFNAMSIZ) {
		fprintf(stderr, "interface name %s too long!\n", if_name);
		exit(1);
	}
	
	strcpy(get_stats.iface.cdnr_ifname, if_name);
	get_stats.nskip = 0;
	get_stats.nelements = NELEMENTS;
	get_stats.tce_stats = elements;

	while (1) {
	
		if (ioctl(cdnrfd, CDNR_GETSTATS, &get_stats) < 0)
			err(1, "ioctl CDNR_GETSTATS");

		gettimeofday(&cur_time, NULL);
		sec = (double)(cur_time.tv_sec - last_time.tv_sec) +
			(double)(cur_time.tv_usec - last_time.tv_usec)
				/ 1000000;

		printf("actions:\n");
		printf("  pass:%llu drop:%llu mark:%llu next:%llu return:%llu none:%llu\n",
		       (ull)get_stats.cnts[TCACODE_PASS].packets,
		       (ull)get_stats.cnts[TCACODE_DROP].packets,
		       (ull)get_stats.cnts[TCACODE_MARK].packets,
		       (ull)get_stats.cnts[TCACODE_NEXT].packets,
		       (ull)get_stats.cnts[TCACODE_RETURN].packets,
		       (ull)get_stats.cnts[TCACODE_NONE].packets);

		for (i = 0; i < get_stats.nelements; i++) {
			tce = &elements[i];
			lp = &last[i];

			if (tce->tce_handle != lp->tce_handle) {
				memset(lp, 0, sizeof(*lp));
				lp->tce_handle = tce->tce_handle;
				continue;
			}

			switch (tce->tce_type) {
			case TCETYPE_TBMETER:
				nprofile = 2;
				profile_names = tbmprof_names;
				break;
			case TCETYPE_TRTCM:
				nprofile = 3;
				profile_names = tcmprof_names;
				break;
			default:
				nprofile = 0;
			}

			if (nprofile == 0)
				continue;

			printf("[%s] handle:%#lx\n",
			       element_names[tce->tce_type], tce->tce_handle);
			for (j = 0; j < nprofile; j++) {
				mbps = (double)(tce->tce_cnts[j].bytes -
						lp->tce_cnts[j].bytes)
					* 8 / sec / 1000000.0; 
				printf("  %s %10llu pkts %16llu bytes (%8.2fMbps)\n",
				       profile_names[j], 
				       tce->tce_cnts[j].packets,
				       tce->tce_cnts[j].bytes, mbps);
			}
			last_time = cur_time;
			*lp = *tce;
		}
		printf("\n");

		sleep(interval);
	}
}
