/*
 * Copyright (C) 1999
 *	Sony Computer Science Laboratories, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: hfscstat.c,v 1.4 1999/12/28 11:57:36 kjc Exp kjc $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <signal.h>
#include <errno.h>

#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/time.h>

#include <sys/socket.h>
#include <net/if.h>

#include <netinet/in.h>
#include <altq/altq.h>
#include <altq/altq_hfsc.h>

#define	HFSC_DEVICE	"/dev/altq/hfsc"
#define NCLASSES	64

int hfscfd = -1;

char *if_name = NULL;
int interval = 5;
int long_format = 0;	/* long format output */
u_quad_t last_bytes[NCLASSES];

int avg_scale = 4096;  /* default fixed-point scale for red */

static void usage(void)
{
	fprintf(stderr, "usage: hfscstat interface [-i interval] [-l]\n");
	exit(1);
}

static void
sigint_handler(int sig)
{
	fprintf(stderr, "Exiting on signal %d\n", sig);

	close(hfscfd);

	exit(0);
}

int main(int argc, char **argv)
{
	struct hfsc_class_stats	get_stats;
	struct class_stats	*sp, *lp, stats[NCLASSES], last[NCLASSES];
	struct timeval		cur_time, last_time;
	int			ch, i;
	double			sec;
	u_int			pkts, drops;
	u_int64_t		total, cumul;

	if (argc < 2)
		usage();

	if_name = argv[1];
	argc--; argv++;

	while ((ch = getopt(argc, argv, "i:l")) != -1) {
		switch (ch) {
		case 'i':
			interval = atoi(optarg);
			break;
		case 'l':
			long_format = 1;
			break;
		default:
			usage();
			break;
		}
	}

	signal(SIGINT, sigint_handler);
	signal(SIGTERM, sigint_handler);
    
	if ((hfscfd = open(HFSC_DEVICE, O_RDONLY)) < 0) {
		perror("hfsc open");
		exit(1);
	}

	if (strlen(if_name) > IFNAMSIZ) {
		fprintf(stderr, "interface name %s too long!\n", if_name);
		exit(1);
	}
	
	strcpy(get_stats.iface.hfsc_ifname, if_name);
	get_stats.stats = &stats[0];

	while (1) {
	
		get_stats.nskip = 0;
		get_stats.nclasses = NCLASSES;
	
		if (ioctl(hfscfd, HFSC_GETSTATS, &get_stats) < 0) {
			perror("ioctl HFSC_GETSTATS");
			exit(1);
		}

		printf("\ncur_time:0x%qx %u classes %u packets in the tree\n",
		       get_stats.cur_time,
		       get_stats.hif_classes, get_stats.hif_packets);

		gettimeofday(&cur_time, NULL);
		sec = (double)(cur_time.tv_sec - last_time.tv_sec) +
			(double)(cur_time.tv_usec - last_time.tv_usec)
				/ 1000000;

		for (i=0; i<get_stats.nclasses; i++) {
			sp = &stats[i];
			lp = &last[i];

			if (sp->class_id != lp->class_id) {
				memset(lp, 0, sizeof(*lp));
				lp->class_id = sp->class_id;
				continue;
			}

			pkts = sp->xmit_cnt.packets - lp->xmit_cnt.packets;
			if (pkts == 0 && long_format == 0)
				continue;
			drops = sp->drop_cnt.packets - lp->drop_cnt.packets;
			total = sp->total - lp->total;
			cumul = sp->cumul - lp->cumul;

			printf("[%2d] handle:0x%lx [rt %uK %ums %uK][ls %uK %ums %uK]\n",
			       sp->class_id, sp->class_handle,
			       sp->rsc.m1/1000, sp->rsc.d, sp->rsc.m2/1000,
			       sp->fsc.m1/1000, sp->fsc.d, sp->fsc.m2/1000);
			printf("  %8.2fMbps [%6.2fM:%6.2fM] qlen:%2d period:%u\n",
			       (double)total * 8.0 / sec / 1000000.0,
			       (double)cumul * 8.0 / sec / 1000000.0,
			       (double)(total - cumul) * 8.0 / sec / 1000000.0,
			       sp->qlength, sp->period);
			printf("     cumul:0x%qx total:0x%qx\n",
			       sp->cumul, sp->total);
			printf("     vt:0x%qx d:0x%qx e:0x%qx\n",
			       sp->vt, sp->d, sp->e);
			if (sp->qtype == Q_RED) {
				printf("     RED q_avg:%.2f xmit: %llu (forced: %u, early: %u marked: %u)\n",
				       ((double)sp->red[0].q_avg)/(double)avg_scale,
				       sp->red[0].xmit_cnt.packets, 
				       sp->red[0].drop_forced,
				       sp->red[0].drop_unforced,
				       sp->red[0].marked_packets);
			}
			else if (sp->qtype == Q_RIO) {
				int dp;

				for (dp = 0; dp < RIO_NDROPPREC; dp++)
					printf("     RIO[%d] q_avg:%.2f xmit: %llu (forced: %u, early: %u marked: %u)\n",
					       dp,
					       ((double)sp->red[dp].q_avg)/(double)avg_scale,
					       sp->red[dp].xmit_cnt.packets, 
					       sp->red[dp].drop_forced,
					       sp->red[dp].drop_unforced,
					       sp->red[dp].marked_packets);
			}

			*lp = *sp;
		}

		last_time = cur_time;
		sleep(interval);
	}
}
